import numpy as np


def rgb2gray(rgb):
    r"""Convert RGB image to Greyscale.

    Parameters
    ----------
    rgb : Numpy array for RGB Image with shape[2] = 3
          The array to convert.

    Returns
    -------
    Numpy array for Greyscale Image.
    """
    return np.dot(rgb[..., :3], [0.299, 0.587, 0.114])


def rgb_to_hsv(arr):
    """
    Borrowed from Matplotlib under the BSD License https://matplotlib.org/devel/license.html

    Convert float rgb values (in the range [0, 1]), in a numpy array to hsv
    values.

    Parameters
    ----------
    arr : (..., 3) array-like
       All values must be in the range [0, 1]

    Returns
    -------
    hsv : (..., 3) ndarray
       Colors converted to hsv values in range [0, 1]
    """
    arr = np.asarray(arr)

    # check length of the last dimension, should be _some_ sort of rgb
    if arr.shape[-1] != 3:
        raise ValueError("Last dimension of input array must be 3; "
                         "shape {} was found.".format(arr.shape))

    in_shape = arr.shape
    arr = np.array(
        arr, copy=False,
        dtype=np.promote_types(arr.dtype, np.float32),  # Don't work on ints.
        ndmin=2,  # In case input was 1D.
    )
    out = np.zeros_like(arr)
    arr_max = arr.max(-1)
    ipos = arr_max > 0
    delta = arr.ptp(-1)
    s = np.zeros_like(delta)
    s[ipos] = delta[ipos] / arr_max[ipos]
    ipos = delta > 0
    # red is max
    idx = (arr[..., 0] == arr_max) & ipos
    out[idx, 0] = (arr[idx, 1] - arr[idx, 2]) / delta[idx]
    # green is max
    idx = (arr[..., 1] == arr_max) & ipos
    out[idx, 0] = 2. + (arr[idx, 2] - arr[idx, 0]) / delta[idx]
    # blue is max
    idx = (arr[..., 2] == arr_max) & ipos
    out[idx, 0] = 4. + (arr[idx, 0] - arr[idx, 1]) / delta[idx]

    out[..., 0] = (out[..., 0] / 6.0) % 1.0
    out[..., 1] = s
    out[..., 2] = arr_max

    return out.reshape(in_shape)



print("import: 'cyvlfeat.fisher.fisher'")
import cyvlfeat.fisher.fisher

print("import: 'cyvlfeat.generic.generic'")
import cyvlfeat.generic.generic

print("import: 'cyvlfeat.hog.hog'")
import cyvlfeat.hog.hog

print("import: 'cyvlfeat.kmeans.kmeans'")
import cyvlfeat.kmeans.kmeans

print("import: 'cyvlfeat.quickshift.quickshift'")
import cyvlfeat.quickshift.quickshift

print("import: 'cyvlfeat.sift.dsift'")
import cyvlfeat.sift.dsift

print("import: 'cyvlfeat.sift.phow'")
import cyvlfeat.sift.phow

print("import: 'cyvlfeat.sift.sift'")
import cyvlfeat.sift.sift

print("import: 'cyvlfeat.vlad.vlad'")
import cyvlfeat.vlad.vlad




import numpy as np
from .cyvlad import cy_vlad


def vlad(x, means, assignments, unnormalized=False, square_root=False,
         normalize_components=False, normalize_mass=False, verbose=False):
    r"""
    Computes the VLAD encoding of the vectors ``x`` relative to
    cluster centers ``means`` and vector-to-cluster soft assignments
    ``assignments``.

    Parameters
    ----------
    x : [D, N]  `float32` `ndarray`
        One column per data vector (e.g. a SIFT descriptor)
    means :  [F, N]  `float32` `ndarray`
        One column per kMeans cluster.
    assignments: [F, N]  `float32` `ndarray`
        No. of rows = No. of clusters
        No. of columns = No. of columns of X
    unnormalized : `bool`, optional
        If ``True``, no overall normalization is applied to the return
        vector.
    square_root : `bool`, optional
        If ``True``, the signed square root function is applied to the return
        vector before normalization.
    normalize_components : `bool`, optional
        If ``True``, the part of the encoding corresponding to each
        cluster is individually normalized.
    normalize_mass : `bool`, optional
        If ``True``, each component is re-normalized by the mass
        of data vectors assigned to it. If ``normalized_components`` is
        also selected, this has no effect.
    verbose: `bool`, optional
        If ``True``, be verbose.

    Returns
    -------
    enc : [k, ] `float32` `ndarray`
        A vector of size equal to the product of
        ``k = the n_data_dimensions * n_clusters``.
        
    Examples:
    --------
    >>> from cyvlfeat.vlad.vlad import vlad
    >>> import numpy as np
    >>> N = 1000
    >>> K = 512
    >>> D = 128
    >>> x = np.random.uniform(size=(D, N)).astype(np.float32)
    >>> means = np.random.uniform(size=(D, K)).astype(np.float32)
    >>> assignments = np.random.uniform(size=(K, N)).astype(np.float32)
    >>> enc = vlad(x, means, assignments)
    """
    # check for None
    if x is None or means is None or assignments is None:
        raise ValueError('A required input is None')

    # validate the KMeans parameters
    D = means.shape[0]  # the feature dimensionality
    K = means.shape[1]  # the number of KMeans centers
    # N = x.shape[1] is the number of samples

    # Check one dimension only.
    if x.shape[0] != D:
        raise ValueError('x and means do not have the same dimensionality')

    if assignments.shape[0] != K:
        raise ValueError('assignments has an unexpected shape')

    result = cy_vlad(x, means, assignments, np.int32(unnormalized),
                     np.int32(square_root), np.int32(normalize_components),
                     np.int32(normalize_mass),
                     np.int32(verbose))

    return result



import numpy as np
from cyvlfeat.quickshift.cyquickshift import cy_quickshift


def quickshift(image, kernel_size, max_dist=None,
               medoid=False, verbose=False):
    r"""
    Quick shift is a mode seeking algorithm which links each pixel to
    its nearest neighbor which has an increase in the estimate of the
    density. These links form a tree, where the root of the tree is
    the pixel which correspond to the highest mode in the image.

    Parameters
    ----------
    image : [H, W] or [H, W, 1] `float64` `ndarray`
        A single channel, greyscale, `float64` numpy array (ndarray).
    kernel_size : ``double``
        The bandwidth parameter for density estimation.
    max_dist : ``double``, optional
        The maximum distance to a neighbor which increases
        the density. Since searching over all pixels for the nearest
        neighbor which increases the density would be prohibitively
        expensive, ``max_dist`` controls the maximum L2 distance between neighbors
        that should be linked.
    medoid : `bool`, optional
        If ``True``, run medoid shift instead of quick shift.
    verbose : `bool`, optional
        If ``True``, be verbose.

    Returns
    -------
    map : [H, W] `float64` `ndarray`. Array of the same size of I.
        Each element (pixel) of ``map`` is an index to the parent
        element in the forest.
    gaps: [H, W] `float64` `ndarray`. Array of the same size of I.
        ``gaps`` contains the corresponding branch length.
        Pixels which are at the root of their respective
        tree have MAP(x) = x and GAPS(x) = inf.
    estimate: [H, W] `float64` `ndarray`, optional
        The estimate of the density. Only returned if
        ``max_dist`` is not ``None``.
        
    Example
    -------
    >>> import numpy as np
    >>> from cyvlfeat.quickshift.quickshift import quickshift
    >>> from cyvlfeat.test_util import lena
    >>> img = lena().astype(np.float32)
    >>> maps, gaps, estimate = quickshift(img,kernel_size=2,max_dist=10)
    >>> # medoid segmentation
    >>> maps, gaps, estimate = quickshift(img,kernel_size=2,max_dist=10,medoid=True)
    """

    # check for None
    if image is None or kernel_size is None:
        raise ValueError('A required input is None')

    # Remove last channel
    if image.ndim == 3 and image.shape[-1] == 1:
        image = image[..., 0]

    # Validate image size
    if image.ndim != 2:
        raise ValueError('Only 2D arrays are supported')

    #     if image.ndim != 3:
    #         raise ValueError('Only 3D arrays are supported')

    compute_estimate = True

    if max_dist is None:
        compute_estimate = False
        # uses a default max_dist of kernel_size * 3
        max_dist = kernel_size * 3

    # Ensure types are correct before passing to Cython
    image = np.require(image, dtype=np.float64, requirements='C')

    result = cy_quickshift(image, kernel_size, max_dist, compute_estimate, medoid, verbose)

    return result



from __future__ import division
from cyvlfeat.sift.dsift import dsift
from cyvlfeat.sift.sift import sift
from cyvlfeat.sift.phow import phow
import numpy as np
from numpy.testing import assert_allclose
from cyvlfeat.test_util import lena


img = lena().astype(np.float32)
half_img = img[:, :256]


def test_phow_float_descriptors():
    i = img.copy()
    frames, descriptors = phow(i, float_descriptors=True)
    assert descriptors.dtype == np.float32


def test_phow_frames():
    i = img.copy()
    frames, descriptors = phow(i, float_descriptors=True)
    assert frames.shape[0] == 236206


def test_phow_descriptors():
    i = img.copy()
    frames, descriptors = phow(i, float_descriptors=True)
    assert descriptors.shape[0] == 236206


def test_dsift_non_float_descriptors():
    i = img.copy()
    frames, descriptors = dsift(i, float_descriptors=False)
    assert descriptors.dtype == np.uint8


def test_dsift_float_descriptors():
    i = img.copy()
    frames, descriptors = dsift(i, float_descriptors=True)
    assert descriptors.dtype == np.float32


def test_dsift_steps():
    i = half_img.copy()
    # Step 3 in Y-Direction, 4 in X-Direction
    frames, descriptors = dsift(i, step=[3, 4])

    assert frames.shape[0] == 10416
    assert_allclose(frames[:3], [[4.5, 4.5], [4.5, 8.5], [4.5, 12.5]],
                    rtol=1e-3)


def test_dsift_windowsize():
    i = half_img.copy()
    frames, descriptors = dsift(i, window_size=3)

    assert frames.shape[0] == 124241
    assert_allclose(frames[:3], [[4.5, 4.5], [4.5, 5.5], [4.5, 6.5]],
                    rtol=1e-3)
    assert_allclose(descriptors[0, -3:], [74, 55, 71],
                    rtol=1e-3)


def test_dsift_fast():
    i = half_img.copy()
    frames, descriptors = dsift(i, fast=True)

    assert frames.shape[0] == 124241
    assert_allclose(frames[:3], [[4.5, 4.5], [4.5, 5.5], [4.5, 6.5]],
                    rtol=1e-3)
    assert_allclose(descriptors[0, -3:], [61, 45, 60],
                    rtol=1e-3)


def test_dsift_norm():
    i = half_img.copy()
    frames, descriptors = dsift(i, norm=True)

    assert frames.shape[-1] == 3
    assert frames.shape[0] == 124241
    assert_allclose(frames[:3], [[4.5, 4.5, 1.6537], [4.5, 5.5, 1.7556],
                                 [4.5, 6.5, 1.8581]],
                    rtol=1e-3)
    assert_allclose(descriptors[0, -3:], [65, 48, 62],
                    rtol=1e-3)


def test_sift_n_frames():
    i = img.copy()
    frames = sift(i)
    assert_allclose(frames[0], [2.16217, 128.056, 2.13029, 5.9325], rtol=1e-3)
    assert frames.shape[0] == 730


def test_sift_non_float_descriptors():
    i = half_img.copy()
    frames, descriptors = sift(i, compute_descriptor=True)

    assert_allclose(frames[0], [2.16217, 128.056, 2.13029, 5.9325], rtol=1e-3)
    assert_allclose(descriptors[0, 84:88], [4,  14,  32, 137])
    assert frames.shape[0] == 358


def test_sift_user_defined_frames():
    i = half_img.copy()
    frames = np.array([[4., 5., 2., 0], [3., 2., 3., np.pi / 2],
                       [8., 6., 4., 1.]])
    new_frames, descriptors = sift(i, frames=frames, compute_descriptor=True)

    assert_allclose(new_frames[0], frames[0], rtol=1e-3)
    assert_allclose(descriptors[0, -5:], [14, 22, 23, 10,  2])
    assert new_frames.shape[0] == 3


def test_sift_sort_user_defined_scales():
    i = half_img.copy()
    frames = np.array([[4., 5., 3., 0], [3., 2., 2., np.pi / 2],
                       [8., 6., 1., 1.]])
    new_frames, descriptors = sift(i, frames=frames, compute_descriptor=True)

    assert_allclose(new_frames[0], frames[-1], rtol=1e-3)
    assert_allclose(descriptors[0, -5:], [46, 14,  0,  0,  0])
    assert new_frames.shape[0] == 3


def test_sift_force_orientations():
    i = half_img.copy()
    frames = np.array([[4., 5., 2., 0], [3., 2., 3., np.pi / 2],
                       [8., 6., 4., 1.]])
    new_frames, descriptors = sift(i, frames=frames, compute_descriptor=True,
                                   force_orientations=True)

    assert_allclose(new_frames[0], [4, 5, 2, 4.6239], rtol=1e-3)
    assert_allclose(descriptors[0, :5], [5, 1, 0, 0, 1])
    assert new_frames.shape[0] == 3



def lena():
    import pickle
    import os
    import numpy as np

    git_repo_path = os.path.join(os.path.dirname(__file__),
                                 '..', 'conda')
    path = os.environ.get('RECIPE_DIR', git_repo_path)
    path = os.path.join(path, 'lena.dat')
    with open(path, 'rb') as f:
        lena = np.array(pickle.load(f))
    return lena



import math

import numpy as np
from scipy.ndimage import filters as scipy_filters

from . import dsift
from ..utils import rgb2gray, rgb_to_hsv


def phow(image, verbose=False, fast=True, sizes=(4, 6, 8, 10), step=2, color='gray',
         float_descriptors=False, magnification=6, window_size=1.5, contrast_threshold=0.005):
    r"""
    Extracts PHOW features from the ``image``. PHOW is simply dense
    SIFT applied at several resolutions.

    `descriptors` has the same format of `sift()` and `dsift()`. `frames[:,1:2]`
    are the x,y coordinates of the center of each descriptor, `frames[:,3]`
    is the contrast of the descriptor, as returned by `dsift()` (for
    colour variant, contrast is computed on the intensity channel).
    `frames[:,4]` is the size of the bin of the descriptor.

    By default, `phow()` computes the gray-scale variant of the descriptor. The
    `color` option can be used to compute the color variant instead.

    Unlike the  Matlab wrapper of vlfeat, the image
    is pre-smoothed at the desired scale level by gaussian filter provided
    by Scipy: ``scipy.ndimage.filters.gaussian_filter``.

    Parameters
    ----------
    image : [H, W] or [H, W, 1] `float32` `ndarray`
        A single channel, greyscale, `float32` numpy array (ndarray)
        representing the image to calculate descriptors for.
    fast : `bool`, optional
        If ``True``, use a piecewise-flat, rather than Gaussian,
        windowing function. While this breaks exact SIFT equivalence,
        in practice is much faster to compute.
    sizes : (`int`, `int`, `int`), optional
        Scales at which the dense SIFT features are extracted. Each
        value is used as bin size for the dsift() function.
    step : `int`, optional
        A SIFT descriptor is extracted every ``step`` pixels. This allows for
        sub-sampling of the image.
    color : `str`, optional
        Choose between 'gray', 'rgb', 'hsv', and 'opponent'.
    float_descriptors : `bool`, optional
        If ``True``, the descriptor are returned in floating point rather than
        integer format.
    magnification : `int`, optional
        Set the descriptor magnification factor. The scale of the keypoint is
        multiplied by this factor to obtain the width (in pixels) of the spatial
        bins. For instance, if there are there are 4 spatial bins along each
        spatial direction, the ``side`` of the descriptor is approximately ``4 *
        magnification``.
    window_size : `int`, optional
        Set the variance of the Gaussian window that determines the
        descriptor support. It is expressed in units of spatial bins.
    contrast_threshold : `float`, optional
        Contrast threshold below which SIFT features are mapped to
        zero. The input image is scaled to have intensity range in [0,1]
        (rather than [0,255]) and this value is compared to the
        descriptor norm as returned by dsift().
    verbose : bool`, optional
        If ``True``, be verbose.

    Returns
    -------
    frames : `(F, 4)` `float32` `ndarray`
        ``F`` is the number of keypoints (frames) used. This is the center
        of every dense SIFT descriptor that is extracted.
    descriptors : `(F, 128)` `uint8` or `float32` `ndarray`
        ``F`` is the number of keypoints (frames) used. The 128 length vectors
        per keypoint extracted. ``uint8`` by default.

    Examples
    --------
      >>> import numpy as np
      >>> from cyvlfeat.sift import phow
      >>> from cyvlfeat.test_util import lena
      >>> img = lena().astype(np.float32)
      >>> frames, descriptors = phow(img, verbose=True)


    """

    # Standardize the image: The following block assumes that the user input
    # for argument color has somewhat more priority than
    # actual color space of I.
    # That is why the conversions are according to the value of variable 'color'
    # irrespective of actual color space to which I belongs.

    color_lower = color.lower()
    I = image.copy()

    # case where user inputs, color ='gray' and I is also greyscale.
    if color_lower == 'gray':
        num_channels = 1

        # case where user inputs, color ='gray' but I belongs to RGB space.
        if I.ndim == 3 and I.shape[2] > 1:
            I = rgb2gray(I)
    else:
        num_channels = 3

        # case where user inputs, color ='rgb'or 'hsv'or 'opponent' but I is greyscale.
        if I.shape[2] == 1:
            I = np.dstack([I, I, I])

        # case where user inputs, color ='rgb' and I also belongs to RGB space.
        elif color_lower == 'rgb':
            pass

        # case where user inputs, color ='opponent' and I belongs to RGB space.
        elif color_lower == 'opponent':

            # Note that the mean differs from the standard definition of opponent
            # space and is the regular intensity (for compatibility with
            # the contrast thresholding).
            # Note also that the mean is added pack to the other two
            # components with a small multipliers for monochromatic
            # regions.

            alpha = 0.01
            I = np.concatenate(
                (rgb2gray(I), (I[:, :, 0] - I[:, :, 1]) / math.sqrt(2) + alpha * rgb2gray(I),
                 I[:, :, 0] + I[:, :, 1] - 2 * I[:, :, 2] / math.sqrt(6) + alpha * rgb2gray(I)),
                axis=2)
        # case when user inputs, color ='hsv' and I belongs to RGB space.
        elif color_lower == 'hsv':
            I = rgb_to_hsv(I)
        else:
            # case when user inputs, color ='hsv' and I belongs to RGB space.
            color_lower = 'hsv'
            I = rgb_to_hsv(I)
            print('Color space not recognized, defaulting to HSV color space.')

    if verbose:
        print('Color space: {}'.format(color))
        print('I size: {}x{}'.format(I.shape[0], I.shape[1]))
        print('Sizes: [{} {} {} {}]'.format(sizes[0], sizes[1], sizes[2], sizes[3]))

    temp_frames = []
    temp_descrs = []

    for si in range(0, len(sizes)):
        f = []
        d = []
        off = math.floor(1.0 + 3.0 / 2.0 * (max(sizes) - sizes[si]))

        # smooth I to the appropriate scale based on the size of the SIFT bins
        sigma = sizes[si] / magnification
        ims = scipy_filters.gaussian_filter(I, sigma)

        # extract dense SIFT features from all channels
        temp_all_results = []
        # temp_arr = np.empty((num_channels, ), dtype=np.float32, order='C')
        data = ims.copy()
        for k in range(num_channels):

            # The third dimension of an image matrix represent the no. of channels that are present.
            # In Matlab, size(I) returns: 256 x256 which is same as the result returned by python's I.shape
            # where I is the numpy array of image. In Matlab, size(I,3) returns 1 for a greyscale
            # image but in Python, I.shape[2] raises an error -> tuple index out of range, simply because
            # there is no third channel. For RGB images I.shape[2] returns 3. The below if-else is a fix
            # for that.
            if ims.ndim == 2:
                # Since it is greyscale, we'd pass whole array (Dsift accepts only 2D arrays.)
                smoothed_image = data

            elif ims.ndim == 3:
                # Since it has 3 channels, i.e. could be split into 3 different channels(2D array) one by one.
                smoothed_image = data[:, :, k]
            else:
                raise ValueError('Image array not defined')

            temp_results = dsift(smoothed_image, step=step, size=sizes[si],
                                 bounds=np.array([off, off, image.shape[0] - 1, image.shape[1] - 1]),
                                 norm=True, fast=fast, float_descriptors=float_descriptors, verbose=verbose)

            temp_all_results.append(temp_results)

        for i in range(len(temp_all_results)):
            f.append(temp_all_results[i][2 * i])
            d.append(temp_all_results[i][2 * i + 1])

        if color_lower == 'gray':
            contrast = f[0][:, 2]
        elif color_lower == 'opponent':
            contrast = f[0][:, 2]
        elif color_lower == 'rgb':
            m = (f[0][:, 2], f[1][:, 2], f[2][:, 2])
            contrast = np.mean(m, axis=1)
        else:
            color_lower = 'hsv'
            contrast = f[2][:, 2]

        # remove low contrast descriptors note that for color descriptors the V component is thresholded

        for k in range(num_channels):
            for i, val in enumerate(contrast):
                if val < contrast_threshold:
                    d[k][i] = 0

        dim2 = contrast.shape[0]
        param2 = (sizes[si]) * np.ones((dim2, 1))
        temp_frames.append(np.append(f[0], param2, axis=1))
        frames = np.concatenate(temp_frames, axis=0)

        temp_descrs.append(d[0])
        descriptors = np.concatenate(temp_descrs, axis=0)

    return frames, descriptors
